@@ -7,7 +7,7 @@
  *
  * Prototypes for functions in commands/comment.c
  *
- * Copyright (c) 1999-2013, PostgreSQL Global Development Group
+ * Copyright (c) 1999-2017, PostgreSQL Global Development Group
  *
  *-------------------------------------------------------------------------
  */
@@ -15,6 +15,7 @@
 #ifndef COMMENT_H
 #define COMMENT_H
 
+#include "catalog/objectaddress.h"
 #include "nodes/parsenodes.h"
 
 /*------------------------------------------------------------------
@@ -24,12 +25,12 @@
  * related routines.  CommentObject() implements the SQL "COMMENT ON"
  * command.  DeleteComments() deletes all comments for an object.
  * CreateComments creates (or deletes, if comment is NULL) a comment
- * for a specific key.	There are versions of these two methods for
+ * for a specific key.  There are versions of these two methods for
  * both normal and shared objects.
  *------------------------------------------------------------------
  */
 
-extern Oid	CommentObject(CommentStmt *stmt);
+extern ObjectAddress CommentObject(CommentStmt *stmt);
 
 extern void DeleteComments(Oid oid, Oid classoid, int32 subid);
 
@@ -41,4 +42,4 @@ extern void CreateSharedComments(Oid oid, Oid classoid, char *comment);
 
 extern char *GetComment(Oid oid, Oid classoid, int32 subid);
 
-#endif   /* COMMENT_H */
+#endif							/* COMMENT_H */
